from .analysis import FunctionArgParser as cmdfunc
from .analysis import UnsupportedTypeException
from .analysis import WrongInputTypeException
from .analysis import NonHomogenousEnumTypeException
from .analysis import EnumHasNoTypeException
from .analysis import IsNotAFunctionException
from .cmdclass import cmdfusion